from typing import Any

import lm_eval
import torch
from lm_eval.models.huggingface import HFLM
from lm_eval.models.openai_completions import OpenaiCompletionsLM

from lm_buddy.integrations.huggingface import (
    AutoModelConfig,
    HuggingFaceAssetLoader,
    HuggingFaceAssetPath,
    resolve_peft_and_pretrained,
)
from lm_buddy.integrations.wandb import (
    ArtifactLoader,
    ArtifactType,
    WandbResumeMode,
    build_table_artifact,
    default_artifact_name,
    wandb_init_from_config,
)
from lm_buddy.jobs.common import LMBuddyJobType
from lm_buddy.jobs.configs import LMHarnessJobConfig, LocalChatCompletionsConfig


def get_numeric_metrics(
    results: dict[str, dict[str, Any]],
) -> dict[str, list[tuple[str, float]]]:
    """Filter non-numeric values from the evaluation results.

    This is necessary because artifact tables must have a single datatype for each column.

    lm-harness returns mostly numeric values, but there are also some misc string-valued metrics.
    Filtering down to only numeric values allows us to produce a valid table artifact.
    """
    numeric_results = {}
    for key, data in results.items():
        numeric_rows = [(k, v) for k, v in data.items() if isinstance(v, int | float)]
        numeric_results[key] = numeric_rows
    return numeric_results


def load_harness_model(
    config: LMHarnessJobConfig,
    artifact_loader: ArtifactLoader,
) -> HFLM | OpenaiCompletionsLM:
    # Instantiate the lm-harness LM class based on the provided model config type
    hf_loader = HuggingFaceAssetLoader(artifact_loader)
    match config.model:
        case AutoModelConfig() as model_config:
            model_path, revision = hf_loader.resolve_asset_path(model_config.load_from)
            model_path, peft_path = resolve_peft_and_pretrained(model_path)
            quantization_kwargs: dict[str, Any] = (
                config.quantization.model_dump() if config.quantization else {}
            )
            # TODO: Fix this up by passing in the instantiated model directly
            return HFLM(
                pretrained=model_path,
                tokenizer=model_path,
                peft=peft_path,
                revision=revision if revision else "main",
                device="cuda" if torch.cuda.device_count() > 0 else "cpu",
                trust_remote_code=config.model.trust_remote_code,
                dtype=config.model.torch_dtype if config.model.torch_dtype else "auto",
                **quantization_kwargs,
            )

        case LocalChatCompletionsConfig() as local_config:
            model = local_config.inference.engine
            if isinstance(model, HuggingFaceAssetPath):
                model, _ = hf_loader.resolve_asset_path(model)
            # If tokenizer is not provided, it is set to the value of model internally
            return OpenaiCompletionsLM(
                model=model,
                base_url=local_config.inference.base_url,
                tokenizer_backend=local_config.tokenizer_backend,
                truncate=local_config.truncate,
                max_gen_toks=local_config.max_tokens,
            )

        case _:
            raise ValueError(f"Unexpected model config type: {type(config.model)}")


def load_and_evaluate(
    config: LMHarnessJobConfig,
    artifact_loader: ArtifactLoader,
) -> dict[str, list[tuple[str, float]]]:
    print("Initializing lm-harness tasks...")
    lm_eval.tasks.initialize_tasks()

    llm = load_harness_model(config, artifact_loader)
    eval_results = lm_eval.simple_evaluate(
        model=llm,
        tasks=config.evaluator.tasks,
        batch_size=config.evaluator.batch_size,
        num_fewshot=config.evaluator.num_fewshot,
        limit=config.evaluator.limit,
        log_samples=False,
    )
    eval_results = get_numeric_metrics(eval_results["results"])
    print(f"Obtained evaluation results: {eval_results}")
    return eval_results


def run_lm_harness(config: LMHarnessJobConfig, artifact_loader: ArtifactLoader):
    print(f"Received job configuration:\n {config.model_dump_json(indent=2)}")

    if config.tracking is not None:
        with wandb_init_from_config(
            config.tracking,
            parameters=config.evaluator,  # Log eval settings in W&B run
            resume=WandbResumeMode.ALLOW,
            job_type=LMBuddyJobType.EVALUATION,
        ) as run:
            eval_results = load_and_evaluate(config, artifact_loader)
            eval_artifact = build_table_artifact(
                artifact_name=default_artifact_name(run.name, ArtifactType.EVALUATION),
                artifact_type=ArtifactType.EVALUATION,
                columns=["metric", "value"],
                tables=eval_results,
            )
            print("Logging artifact for evaluation results...")
            artifact_loader.log_artifact(eval_artifact)
    else:
        load_and_evaluate(config, artifact_loader)



from typing import Literal

from pydantic import conlist, model_validator

from lm_buddy.integrations.huggingface import (
    AutoModelConfig,
    QuantizationConfig,
)
from lm_buddy.integrations.vllm import InferenceServerConfig
from lm_buddy.integrations.wandb import WandbRunConfig
from lm_buddy.jobs.configs import LMBuddyJobConfig
from lm_buddy.types import BaseLMBuddyConfig


class LocalChatCompletionsConfig(BaseLMBuddyConfig):
    """Configuration for a "local-chat-completions" model in lm-harness.

    The "local-chat-completions" model is powered by a self-hosted inference server,
    specified as an `InferenceServerConfig`. Additional arguments are also provided
    to control the tokenizer type and generation parameters.
    """

    inference: InferenceServerConfig
    truncate: bool = False
    max_tokens: int = 256
    tokenizer_backend: Literal["huggingface", "tiktoken"] = "huggingface"

    @model_validator(mode="after")
    def validate_inference_engine(cls, config: "LocalChatCompletionsConfig"):
        if config.inference.engine is None:
            raise ValueError(
                "Inference config `engine` must be provided for use in "
                "lm-harness 'local-chat-completions' model."
            )
        return config


class LMHarnessEvaluatorConfig(BaseLMBuddyConfig):
    """Misc settings provided to an lm-harness evaluation job."""

    tasks: conlist(str, min_length=1)
    batch_size: int | None = None
    num_fewshot: int | None = None
    limit: int | float | None = None


class LMHarnessJobConfig(LMBuddyJobConfig):
    """Configuration to run an lm-evaluation-harness evaluation job."""

    model: AutoModelConfig | LocalChatCompletionsConfig
    evaluator: LMHarnessEvaluatorConfig
    quantization: QuantizationConfig | None = None
    tracking: WandbRunConfig | None = None



from lm_buddy.integrations.huggingface import HuggingFaceAssetPath
from lm_buddy.types import BaseLMBuddyConfig


class InferenceServerConfig(BaseLMBuddyConfig):
    """Generic configuration for an externally hosted inference endpoint.

    The inference server is defined by an endpoint with the provided `base_url`.
    The model `engine` powering the endpoint can also be specified
    to provide further metadata about the model used for inference
    (e.g., for generating W&B artifact lineages).

    Note: This configuration is intended to be generic and not bound to the interface
    of any specific training/evaluation framework. See `LocalChatCompletionConfig`
    for intended usage alongside a third-party framework.
    """

    base_url: str
    engine: str | HuggingFaceAssetPath | None = None



from transformers import BitsAndBytesConfig

from lm_buddy.types import BaseLMBuddyConfig, SerializableTorchDtype


class QuantizationConfig(BaseLMBuddyConfig):
    """Basic quantization settings to pass to training and evaluation jobs.

    Note that in order to use BitsAndBytes quantization on Ray,
    you must ensure that the runtime environment is installed with GPU support.
    This can be configured by setting the `entrypoint_num_gpus > 0` when submitting a job
    to the cluster.
    """

    load_in_8bit: bool | None = None
    load_in_4bit: bool | None = None
    bnb_4bit_quant_type: str = "fp4"
    bnb_4bit_compute_dtype: SerializableTorchDtype | None = None

    def as_huggingface(self) -> BitsAndBytesConfig:
        return BitsAndBytesConfig(
            load_in_4bit=self.load_in_4bit,
            load_in_8bit=self.load_in_8bit,
            bnb_4bit_compute_dtype=self.bnb_4bit_compute_dtype,
            bnb_4bit_quant_type=self.bnb_4bit_quant_type,
        )



import warnings

import torch
from accelerate import Accelerator
from datasets import Dataset, DatasetDict, load_dataset, load_from_disk
from peft import PeftConfig
from transformers import (
    AutoConfig,
    AutoModelForCausalLM,
    AutoTokenizer,
    PretrainedConfig,
    PreTrainedModel,
    PreTrainedTokenizer,
)

from lm_buddy.integrations.huggingface import (
    AutoModelConfig,
    AutoTokenizerConfig,
    DatasetConfig,
    HuggingFaceRepoConfig,
    QuantizationConfig,
)
from lm_buddy.integrations.wandb import (
    ArtifactLoader,
    WandbArtifactConfig,
    get_artifact_filesystem_path,
)

HuggingFaceAssetPath = HuggingFaceRepoConfig | WandbArtifactConfig
"""Config that can be resolved to a HuggingFace name/path."""


def resolve_peft_and_pretrained(path: str) -> tuple[str, str | None]:
    """Helper method for determining if a path corresponds to a PEFT model.

    A PEFT model contains an `adapter_config.json` in its directory.
    If this file can be loaded, we know the path is a for a PEFT model.
    If not, we assume the provided path corresponds to a base HF model.

    Args:
        path (str): Name/path to a HuggingFace directory

    Returns:
        Tuple of (base model path, optional PEFT path)
    """
    # We don't know if the checkpoint is adapter weights or merged model weights
    # Try to load as an adapter and fall back to the checkpoint containing the full model
    try:
        peft_config = PeftConfig.from_pretrained(path)
        return peft_config.base_model_name_or_path, path
    except ValueError as e:
        warnings.warn(
            f"Unable to load model as adapter: {e}. "
            "This is expected if the checkpoint does not contain adapter weights."
        )
        return path, None


class HuggingFaceAssetLoader:
    """Helper class for loading HuggingFace assets from LM Buddy configurations.

    This class depends on an `ArtifactLoader` in order to resolve actual paths from
    artifact references.
    """

    def __init__(self, artifact_loader: ArtifactLoader):
        self._artifact_loader = artifact_loader

    def resolve_asset_path(self, path: HuggingFaceAssetPath) -> tuple[str, str | None]:
        """Resolve the actual HuggingFace name/path from a config.

        Currently, two config types contain references to a loadable HuggingFace path:
        (1) A `HuggingFaceRepoConfig` that contains the repo path directly
        (2) A `WandbArtifactConfig` where the filesystem path is resolved from the artifact
        """
        match path:
            case HuggingFaceRepoConfig(repo_id, revision):
                load_path, revision = repo_id, revision
            case WandbArtifactConfig() as artifact_config:
                artifact = self._artifact_loader.use_artifact(artifact_config)
                load_path = get_artifact_filesystem_path(artifact)
                revision = None
            case unknown_path:
                raise ValueError(f"Unable to resolve asset path from {unknown_path}.")
        return str(load_path), revision

    def load_pretrained_config(
        self,
        config: AutoModelConfig,
    ) -> PretrainedConfig:
        """Load a `PretrainedConfig` from the model configuration.

        An exception is raised if the HuggingFace repo does not contain a `config.json` file.
        """
        model_path, revision = self.resolve_asset_path(config.load_from)
        return AutoConfig.from_pretrained(
            pretrained_model_name_or_path=model_path, revision=revision
        )

    def load_pretrained_model(
        self,
        config: AutoModelConfig,
        quantization: QuantizationConfig | None = None,
    ) -> PreTrainedModel:
        """Load a `PreTrainedModel` with optional quantization from the model configuration.

        An exception is raised if the HuggingFace repo does not contain a `config.json` file.

        TODO(RD2024-87): This fails if the checkpoint only contains a PEFT adapter config
        """
        device_map, bnb_config = None, None
        if quantization is not None:
            bnb_config = quantization.as_huggingface()
            # When quantization is enabled, model must all be on same GPU to work with DDP
            # If a device_map is not specified we will get accelerate errors downstream
            # Reference: https://github.com/huggingface/accelerate/issues/1840#issuecomment-1683105994
            current_device = (
                Accelerator().local_process_index if torch.cuda.is_available() else "cpu"
            )
            device_map = {"": current_device}
            print(f"Setting model device_map = {device_map} to enable quantization")

        # TODO: HuggingFace has many AutoModel classes with different "language model heads"
        #   Can we abstract this to load with any type of AutoModel class?
        model_path, revision = self.resolve_asset_path(config.load_from)
        return AutoModelForCausalLM.from_pretrained(
            pretrained_model_name_or_path=model_path,
            revision=revision,
            trust_remote_code=config.trust_remote_code,
            torch_dtype=config.torch_dtype,
            quantization_config=bnb_config,
            device_map=device_map,
        )

    def load_pretrained_tokenizer(self, config: AutoTokenizerConfig) -> PreTrainedTokenizer:
        """Load a `PreTrainedTokenizer` from the model configuration.

        An exception is raised if the HuggingFace repo does not contain a `tokenizer.json` file.
        """
        tokenizer_path, revision = self.resolve_asset_path(config.load_from)
        tokenizer = AutoTokenizer.from_pretrained(
            pretrained_model_name_or_path=tokenizer_path,
            revision=revision,
            trust_remote_code=config.trust_remote_code,
            use_fast=config.use_fast,
        )
        if tokenizer.pad_token_id is None:
            # Pad token required for generating consistent batch sizes
            tokenizer.pad_token_id = tokenizer.eos_token_id
        return tokenizer

    def load_dataset(self, config: DatasetConfig) -> Dataset:
        """Load a HuggingFace `Dataset` from the dataset configuration.

        This method always returns a single `Dataset` object.
        When loading from HuggingFace directly, the `Dataset` is for the provided split.
        When loading from disk, the saved files must be for a dataset else an exception is raised.
        """
        dataset_path, revision = self.resolve_asset_path(config.load_from)
        # Dataset loading requires a different method if from a HF vs. disk
        if isinstance(config.load_from, HuggingFaceRepoConfig):
            return load_dataset(dataset_path, revision=revision, split=config.split)
        else:
            match load_from_disk(dataset_path):
                case Dataset() as dataset:
                    return dataset
                case other:
                    raise ValueError(
                        "LM Buddy currently only supports loading `Dataset` objects from disk, "
                        f"instead found a {type(other)}."
                    )

    def load_and_split_dataset(self, config: DatasetConfig) -> DatasetDict:
        """Load a HuggingFace dataset and optionally perform a train/test split.

        The split is performed when a `test_size` is specified on the configuration.
        """
        match self.load_dataset(config):
            case Dataset() as dataset if config.test_size is not None:
                # We need to specify a fixed seed to load the datasets on each worker
                # Under the hood, HuggingFace uses `accelerate` to create a data loader shards
                # If the datasets are not seeded here, the ordering will be inconsistent
                # TODO: Get rid of this when data is loaded once outside of Ray workers
                split_seed = config.seed or 0
                return dataset.train_test_split(test_size=config.test_size, seed=split_seed)
            case dataset:
                return DatasetDict({"train": dataset})



from enum import Enum


class LMBuddyJobType(str, Enum):
    """Enumeration of logical job types runnable via the LM Buddy."""

    PREPROCESSING = "preprocessing"
    FINETUNING = "finetuning"
    EVALUATION = "evaluation"



from pydantic import field_validator, model_validator

from lm_buddy.integrations.huggingface import HuggingFaceRepoConfig, convert_string_to_repo_config
from lm_buddy.integrations.wandb import WandbArtifactConfig
from lm_buddy.types import BaseLMBuddyConfig

DEFAULT_TEXT_FIELD: str = "text"


class DatasetConfig(BaseLMBuddyConfig):
    """Base configuration to load a HuggingFace dataset."""

    load_from: HuggingFaceRepoConfig | WandbArtifactConfig
    split: str | None = None
    test_size: float | None = None
    seed: int | None = None

    _validate_load_from_string = field_validator("load_from", mode="before")(
        convert_string_to_repo_config
    )

    @model_validator(mode="after")
    def validate_split_if_huggingface_repo(cls, config: "DatasetConfig"):
        """
        Ensure a  `split` is provided when loading a HuggingFace dataset directly from HF Hub.
        This makes it such that the `load_dataset` function returns the type `Dataset`
        instead of `DatasetDict`, which makes some of the downstream logic easier.
        """
        load_from = config.load_from
        split = config.split
        if split is None and isinstance(load_from, HuggingFaceRepoConfig):
            raise ValueError(
                "A `split` must be specified when loading a dataset directly from HuggingFace."
            )
        return config


class TextDatasetConfig(DatasetConfig):
    """Settings passed to load a HuggingFace text dataset.

    Inherits fields from the the base `DatasetConfig`.
    The dataset should contain a single text column named by the `text_field` parameter.
    """

    text_field: str = DEFAULT_TEXT_FIELD



import contextlib
import tempfile
from pathlib import Path

from pydantic_yaml import parse_yaml_file_as, to_yaml_file

from lm_buddy.types import BaseLMBuddyConfig


class LMBuddyJobConfig(BaseLMBuddyConfig):
    """Configuration that comprises the entire input to an LM Buddy job.

    This class implements helper methods for de/serializing the configuration from file.

    Currently, there is a 1:1 mapping between job entrypoints and job config implementations,
    but this is not rigidly constrained by the interface. This may change in the future.
    """

    @classmethod
    def from_yaml_file(cls, path: Path | str):
        return parse_yaml_file_as(cls, path)

    def to_yaml_file(self, path: Path | str):
        to_yaml_file(path, self, exclude_none=True)

    @contextlib.contextmanager
    def to_tempfile(self, *, name: str = "config.yaml", dir: str | Path | None = None):
        """Enter a context manager with the config written to a temporary YAML file.

        Keyword Args:
            name (str): Name of the config file in the tmp directory. Defaults to "config.yaml".
            dir (str | Path | None): Root path of the temporary directory. Defaults to None.

        Returns:
            Path to the temporary config file.
        """
        with tempfile.TemporaryDirectory(dir=dir) as tmpdir:
            config_path = Path(tmpdir) / name
            self.to_yaml_file(config_path)
            yield config_path



import re

from lm_buddy.types import BaseLMBuddyConfig


class WandbArtifactConfig(BaseLMBuddyConfig):
    """Configuration required to retrieve an artifact from W&B."""

    name: str
    project: str
    version: str = "latest"
    entity: str | None = None

    @classmethod
    def from_wandb_path(cls, path: str) -> "WandbArtifactConfig":
        """Construct an artifact configuration from the W&B name.

        The name should be of the form "<entity>/<project>/<name>:<version>"
        with the "entity" field optional.
        """
        match = re.search(r"((.*)\/)?(.*)\/(.*)\:(.*)", path)
        if match is not None:
            entity, project, name, version = match.groups()[1:]
            return cls(name=name, project=project, version=version, entity=entity)
        raise ValueError(f"Invalid artifact path: {path}")

    def wandb_path(self) -> str:
        """String identifier for the asset on the W&B platform."""
        path = f"{self.project}/{self.name}:{self.version}"
        if self.entity is not None:
            path = f"{self.entity}/{path}"
        return path



from pydantic import field_validator

from lm_buddy.integrations.huggingface import HuggingFaceRepoConfig, convert_string_to_repo_config
from lm_buddy.integrations.wandb import WandbArtifactConfig
from lm_buddy.types import BaseLMBuddyConfig, SerializableTorchDtype


class AutoModelConfig(BaseLMBuddyConfig):
    """Settings passed to a HuggingFace AutoModel instantiation.

    The model to load can either be a HuggingFace repo or an artifact reference on W&B.
    """

    load_from: HuggingFaceRepoConfig | WandbArtifactConfig
    trust_remote_code: bool = False
    torch_dtype: SerializableTorchDtype | None = None

    _validate_load_from_string = field_validator("load_from", mode="before")(
        convert_string_to_repo_config
    )



import wandb
from pydantic import BaseModel

from lm_buddy.integrations.wandb import WandbArtifactConfig


def copy_pydantic_json(model: BaseModel) -> BaseModel:
    """Copy a Pydantic model through round-trip JSON serialization."""
    return model.__class__.model_validate_json(model.model_dump_json())


class FakeArtifactLoader:
    """Fake implementation of an `ArtifactLoader` with in-memory artifact storage.

    This class bypasses calls to the W&B SDK for using/logging artifacts,
    making it suitable for use in testing when W&B is disabled.

    Note: Artifacts are retrieved from the in-memory storage using just their name,
    not the full W&B path, since the project/entity cannot be inferred when W&B is disabled.
    """

    def __init__(self):
        self._storage: dict[str, wandb.Artifact] = dict()

    def num_artifacts(self) -> int:
        return len(self._storage)

    def get_artifacts(self) -> list[wandb.Artifact]:
        return list(self._storage.values())

    def use_artifact(self, config: WandbArtifactConfig) -> wandb.Artifact:
        return self._storage[config.name]

    def log_artifact(self, artifact: wandb.Artifact) -> None:
        self._storage[artifact.name] = artifact



import json

import click

from lm_buddy.jobs.configs import FinetuningJobConfig, LMHarnessJobConfig, SimpleJobConfig


@click.group(name="schema", help="Get a job configuration schema.")
def group():
    pass


@group.command("simple", help="Schema for the simple test job configuration.")
def schema_simple() -> None:
    schema = SimpleJobConfig.model_json_schema()
    click.secho(json.dumps(schema, indent=2))


@group.command("finetuning", help="Schema for the finetuning job configuration.")
def schema_finetuning() -> None:
    schema = FinetuningJobConfig.model_json_schema()
    click.secho(json.dumps(schema, indent=2))


@group.command("lm-harness", help="Schema for the lm-harness job configuration.")
def schema_lm_harness() -> None:
    schema = LMHarnessJobConfig.model_json_schema()
    click.secho(json.dumps(schema, indent=2))



from lm_buddy.integrations.wandb import ArtifactLoader, WandbArtifactLoader
from lm_buddy.jobs._entrypoints import run_finetuning, run_lm_harness, run_simple
from lm_buddy.jobs.configs import (
    FinetuningJobConfig,
    LMBuddyJobConfig,
    LMHarnessJobConfig,
    SimpleJobConfig,
)


def run_job(
    config: LMBuddyJobConfig,
    artifact_loader: ArtifactLoader = WandbArtifactLoader(),
) -> None:
    """Run an LM Buddy job from the configuration.

    Args:
        config (LMBuddyJobConfig): Configuration defining the job to run.
        artifact_loader (ArtifactLoader): Implementation of the artifact loader protocol.
            Defaults to WandbArtifactLoader().
    """
    match config:
        case SimpleJobConfig() as simple_config:
            run_simple(simple_config, artifact_loader)
        case FinetuningJobConfig() as finetuning_config:
            run_finetuning(finetuning_config, artifact_loader)
        case LMHarnessJobConfig() as lm_harness_config:
            run_lm_harness(lm_harness_config, artifact_loader)
        case _:
            raise ValueError(f"Received invalid job configuration: {config}")



import dataclasses

from peft import PeftConfig, PeftType, TaskType
from pydantic import field_validator, model_validator

from lm_buddy.types import BaseLMBuddyConfig


def _get_peft_config_class(peft_type: PeftType) -> type[PeftConfig]:
    # Internal import to avoid bringing the global variable from peft into module scope
    from peft.mapping import PEFT_TYPE_TO_CONFIG_MAPPING

    return PEFT_TYPE_TO_CONFIG_MAPPING[peft_type]


class AdapterConfig(BaseLMBuddyConfig, extra="allow"):
    """Configuration containing PEFT adapter settings.

    The type of adapter is controlled by the required `peft_type` field,
    which must be one of the allowed values from the PEFT `PeftType` enumeration.
    Extra arguments are allowed and are passed down to the HuggingFace `PeftConfig`
    class determined by the `peft_type` argument.

    The `task_type` for the adapter is also required.
    By default, this is set to `TaskType.CAUSAL_LM`
    which is appropriate for causal language model finetuning.
    See the allowed values in the PEFT `TaskType` enumeration.
    """

    peft_type: PeftType
    task_type: TaskType = TaskType.CAUSAL_LM

    @field_validator("peft_type", "task_type", mode="before")
    def sanitize_enum_args(cls, x):
        if isinstance(x, str):
            x = x.strip().upper()
        return x

    @model_validator(mode="after")
    def validate_adapter_args(cls, config: "AdapterConfig"):
        peft_type = config.peft_type

        # PeftConfigs are standard dataclasses so can extract their allowed field names
        adapter_cls = _get_peft_config_class(peft_type)
        allowed_fields = {x.name for x in dataclasses.fields(adapter_cls)}

        # Filter fields to those found on the PeftConfig
        extra_fields = config.model_fields_set.difference(allowed_fields)
        if extra_fields:
            raise ValueError(f"Unknowon arguments for {peft_type} adapter: {extra_fields}")

        return config

    def as_huggingface(self) -> PeftConfig:
        adapter_cls = _get_peft_config_class(self.peft_type)
        return adapter_cls(**self.model_dump())



import click

from lm_buddy.cli import run, schema


@click.group(name="LM Buddy CLI", help="Entrypoints for the LM Buddy.")
def cli():
    pass


cli.add_command(run.group)
cli.add_command(schema.group)



from typing import Any

from huggingface_hub.utils import HFValidationError, validate_repo_id
from pydantic import field_validator

from lm_buddy.types import BaseLMBuddyConfig


def convert_string_to_repo_config(x: Any):
    if isinstance(x, str):
        return HuggingFaceRepoConfig(repo_id=x)
    return x


def is_valid_huggingface_repo_id(s: str):
    """
    Simple test to check if an HF model is valid using HuggingFace's tools.
    Sadly, theirs throws an exception and has no return.

    Args:
        s: string to test.
    """
    try:
        validate_repo_id(s)
        return True
    except HFValidationError:
        return False


class HuggingFaceRepoConfig(BaseLMBuddyConfig):
    """Configuration for a HuggingFace Hub repository."""

    __match_args__ = ("repo_id", "revision")

    repo_id: str
    revision: str | None = None

    @field_validator("repo_id", mode="after")
    def validate_repo_id(cls, x: str):
        if not is_valid_huggingface_repo_id(x):
            raise ValueError(f"{x} is not a valid HuggingFace repo ID.")
        return x



from lm_buddy.jobs.configs.base import LMBuddyJobConfig
from lm_buddy.jobs.configs.finetuning import FinetuningJobConfig, FinetuningRayConfig
from lm_buddy.jobs.configs.lm_harness import (
    LMHarnessEvaluatorConfig,
    LMHarnessJobConfig,
    LocalChatCompletionsConfig,
)
from lm_buddy.jobs.configs.simple import SimpleJobConfig

__all__ = [
    "LMBuddyJobConfig",
    "SimpleJobConfig",
    "FinetuningJobConfig",
    "FinetuningRayConfig",
    "LMHarnessEvaluatorConfig",
    "LMHarnessJobConfig",
    "LocalChatCompletionsConfig",
    "SimpleJobConfig",
]



from typing import Any

from lm_buddy.types import BaseLMBuddyConfig


class TrainerConfig(BaseLMBuddyConfig):
    """Configuration for a HuggingFace trainer/training arguments.

    This mainly encompasses arguments passed to the HuggingFace `TrainingArguments` class,
    but also contains some additional parameters for the `Trainer` or `SFTTrainer` classes.
    """

    max_seq_length: int | None = None
    num_train_epochs: float | None = None
    per_device_train_batch_size: int | None = None
    per_device_eval_batch_size: int | None = None
    learning_rate: float | None = None
    weight_decay: float | None = None
    gradient_accumulation_steps: int | None = None
    gradient_checkpointing: bool | None = None
    evaluation_strategy: str | None = None
    eval_steps: float | None = None
    logging_strategy: str | None = None
    logging_steps: float | None = None
    save_strategy: str | None = None
    save_steps: int | None = None

    def training_args(self) -> dict[str, Any]:
        """Return the arguments to the HuggingFace `TrainingArguments` class."""
        return self.model_dump(exclude={"max_seq_length"}, exclude_none=True)



from pydantic import field_validator

from lm_buddy.integrations.huggingface import HuggingFaceRepoConfig, convert_string_to_repo_config
from lm_buddy.integrations.wandb import WandbArtifactConfig
from lm_buddy.types import BaseLMBuddyConfig


class AutoTokenizerConfig(BaseLMBuddyConfig):
    """Settings passed to a HuggingFace AutoTokenizer instantiation."""

    load_from: HuggingFaceRepoConfig | WandbArtifactConfig
    trust_remote_code: bool | None = None
    use_fast: bool | None = None

    _validate_load_from_string = field_validator("load_from", mode="before")(
        convert_string_to_repo_config
    )



from lm_buddy.jobs._entrypoints.finetuning import run_finetuning
from lm_buddy.jobs._entrypoints.lm_harness import run_lm_harness
from lm_buddy.jobs._entrypoints.simple import run_simple

__all__ = ["run_finetuning", "run_lm_harness", "run_simple"]



from lm_buddy.integrations.wandb.artifact_config import *
from lm_buddy.integrations.wandb.artifact_loader import *
from lm_buddy.integrations.wandb.artifact_utils import *
from lm_buddy.integrations.wandb.run_config import *
from lm_buddy.integrations.wandb.run_utils import *



import ray

from lm_buddy.jobs.configs import SimpleJobConfig


@ray.remote
def get_magic_number(config: SimpleJobConfig) -> int:
    return config.magic_number


def run_simple(config: SimpleJobConfig):
    """A simple entrypoint to demonstrate the Ray interface."""
    # Connect to the Ray cluster (if not already running)
    ray.init(ignore_reinit_error=True)
    # Run dummy remote task
    magic_number = ray.get(get_magic_number.remote(config))
    print(f"The magic number is {magic_number}")



import contextlib
from enum import Enum
from typing import Any

import wandb
from wandb.apis.public import Run as ApiRun

from lm_buddy.integrations.wandb import WandbRunConfig
from lm_buddy.types import BaseLMBuddyConfig


class WandbResumeMode(str, Enum):
    """Enumeration of modes for resuming a W&B run.

    This is not an exahustive list of the values that can be passed to the W&B SDK
    (Docs: https://docs.wandb.ai/ref/python/init), but just those commonly used within the package.
    """

    ALLOW = "allow"
    MUST = "must"
    NEVER = "never"


@contextlib.contextmanager
def wandb_init_from_config(
    config: WandbRunConfig,
    *,
    parameters: BaseLMBuddyConfig | None = None,
    resume: WandbResumeMode | None = None,
    job_type: str | None = None,
):
    """Initialize a W&B run from the internal run configuration.

    This method can be entered as a context manager similar to `wandb.init` as follows:

    ```
    with wandb_init_from_config(run_config, resume=WandbResumeMode.MUST) as run:
        # Use the initialized run here
        ...
    ```
    """
    init_kwargs = dict(
        id=config.run_id,
        name=config.name,
        project=config.project,
        entity=config.entity,
        group=config.run_group,
        config=parameters.model_dump(mode="json") if parameters else None,
        job_type=job_type,
        resume=resume,
    )
    with wandb.init(**init_kwargs) as run:
        yield run


def get_wandb_api_run(config: WandbRunConfig) -> ApiRun:
    """Retrieve a run from the W&B API."""
    api = wandb.Api()
    return api.run(config.wandb_path())


def get_wandb_summary(config: WandbRunConfig) -> dict[str, Any]:
    """Get the summary dictionary attached to a W&B run."""
    run = get_wandb_api_run(config)
    return dict(run.summary)


def update_wandb_summary(config: WandbRunConfig, metrics: dict[str, Any]) -> None:
    """Update a run's summary with the provided metrics."""
    run = get_wandb_api_run(config)
    run.summary.update(metrics)
    run.update()



# ruff: noqa: I001
from lm_buddy.integrations.huggingface.repo_config import *
from lm_buddy.integrations.huggingface.adapter_config import *
from lm_buddy.integrations.huggingface.dataset_config import *
from lm_buddy.integrations.huggingface.model_config import *
from lm_buddy.integrations.huggingface.quantization_config import *
from lm_buddy.integrations.huggingface.tokenizer_config import *
from lm_buddy.integrations.huggingface.trainer_config import *
from lm_buddy.integrations.huggingface.asset_loader import *



import click

import lm_buddy
from lm_buddy.jobs.configs import FinetuningJobConfig, LMHarnessJobConfig, SimpleJobConfig

# TODO(RD2024-125): We should probably collapse all these commands into a single CLI command
# - Need to figure out best way to polymorphically deserialize the job config classes
# - Do we just add type discriminators at the job config level?


@click.group(name="run", help="Run an LM Buddy job.")
def group():
    pass


@group.command("simple", help="Run the simple test job.")
@click.option("--config", type=str)
def run_simple(config: str) -> None:
    config = SimpleJobConfig.from_yaml_file(config)
    lm_buddy.run_job(config)


@group.command("finetuning", help="Run the HuggingFace LLM finetuning job.")
@click.option("--config", type=str)
def run_finetuning(config: str) -> None:
    config = FinetuningJobConfig.from_yaml_file(config)
    lm_buddy.run_job(config)


@group.command("lm-harness", help="Run the lm-harness evaluation job.")
@click.option("--config", type=str)
def run_lm_harness(config: str) -> None:
    config = LMHarnessJobConfig.from_yaml_file(config)
    lm_buddy.run_job(config)



from typing import Protocol

import wandb

from lm_buddy.integrations.wandb import WandbArtifactConfig


class ArtifactLoader(Protocol):
    """Base interface for using/logging artifacts.

    Note: If/when we decide to support other tracking services (e.g., MLFlow, CometML),
    this interface should be abstracted to handle the types for those respective services.
    """

    def use_artifact(self, config: WandbArtifactConfig) -> wandb.Artifact:
        """Load an artifact from its configuration.

        If a W&B run is active, the artifact is declared as an input to the run.
        If not, the artifact is retrieved outside of the run.
        """
        pass

    def log_artifact(self, artifact: wandb.Artifact) -> None:
        """Log an artifact, declaring it as an output of the currently active W&B run."""
        pass


class WandbArtifactLoader:
    """Weights & Biases implementation of the `ArtifactLoader` protocol.

    This class makes external calls to the W&B API and is not suitable for test environments.
    """

    def use_artifact(self, config: WandbArtifactConfig) -> wandb.Artifact:
        if wandb.run is not None:
            # Retrieves the artifact and links it as an input to the run
            return wandb.use_artifact(config.wandb_path())
        else:
            # Retrieves the artifact outside of the run
            api = wandb.Api()
            return api.artifact(config.wandb_path())

    def log_artifact(self, artifact: wandb.Artifact) -> None:
        return wandb.log_artifact(artifact)



from typing import Annotated, Any

import torch
from pydantic import BaseModel, BeforeValidator, PlainSerializer, WithJsonSchema


def validate_torch_dtype(x: Any) -> torch.dtype:
    match x:
        case torch.dtype():
            return x
        case str() if hasattr(torch, x) and isinstance(getattr(torch, x), torch.dtype):
            return getattr(torch, x)
        case _:
            raise ValueError(f"{x} is not a valid torch.dtype.")


SerializableTorchDtype = Annotated[
    torch.dtype,
    BeforeValidator(lambda x: validate_torch_dtype(x)),
    PlainSerializer(lambda x: str(x).split(".")[1]),
    WithJsonSchema({"type": "string"}, mode="validation"),
    WithJsonSchema({"type": "string"}, mode="serialization"),
]
"""Custom type validator for a `torch.dtype` object.

Accepts `torch.dtype` instances or strings representing a valid dtype from the `torch` package.
Ref: https://docs.pydantic.dev/latest/concepts/types/#custom-types
"""


class BaseLMBuddyConfig(
    BaseModel,
    extra="forbid",
    arbitrary_types_allowed=True,
    validate_assignment=True,
):
    """Base class for all Pydantic configs in the library.

    Defines some common settings used by all subclasses.
    """



from typing import Any

import ray
from ray import train
from ray.train import CheckpointConfig, RunConfig, ScalingConfig
from ray.train.huggingface.transformers import RayTrainReportCallback, prepare_trainer
from ray.train.torch import TorchTrainer
from transformers import TrainingArguments
from trl import SFTTrainer

from lm_buddy.integrations.huggingface import HuggingFaceAssetLoader
from lm_buddy.integrations.wandb import (
    ArtifactLoader,
    ArtifactType,
    WandbResumeMode,
    build_directory_artifact,
    default_artifact_name,
    wandb_init_from_config,
)
from lm_buddy.jobs.common import LMBuddyJobType
from lm_buddy.jobs.configs import FinetuningJobConfig


def is_tracking_enabled(config: FinetuningJobConfig):
    # Only report to WandB on the rank 0 worker
    # Reference: https://docs.ray.io/en/latest/train/user-guides/experiment-tracking.html
    return config.tracking is not None and train.get_context().get_world_rank() == 0


def load_and_train(config: FinetuningJobConfig, artifact_loader: ArtifactLoader):
    # Load the HF assets from configurations
    # Internally, artifact lineages are declared for the active training run
    hf_loader = HuggingFaceAssetLoader(artifact_loader)
    model = hf_loader.load_pretrained_model(config.model, config.quantization)
    tokenizer = hf_loader.load_pretrained_tokenizer(config.tokenizer)
    datasets = hf_loader.load_and_split_dataset(config.dataset)

    training_args = TrainingArguments(
        output_dir="out",  # Local checkpoint path on a worker
        report_to="wandb" if is_tracking_enabled(config) else "none",
        use_cpu=not config.ray.use_gpu,
        push_to_hub=False,
        disable_tqdm=True,
        logging_dir=None,
        **config.trainer.training_args(),
    )

    peft_config = config.adapter.as_huggingface() if config.adapter else None

    trainer = SFTTrainer(
        model=model,
        tokenizer=tokenizer,
        args=training_args,
        peft_config=peft_config,
        train_dataset=datasets["train"],
        eval_dataset=datasets.get("test"),
        max_seq_length=config.trainer.max_seq_length,
        dataset_text_field=config.dataset.text_field,
    )
    trainer.add_callback(RayTrainReportCallback())
    trainer = prepare_trainer(trainer)
    trainer.train()


def run_finetuning(config: FinetuningJobConfig, artifact_loader: ArtifactLoader):
    # Place the artifact loader in Ray object store
    artifact_loader_ref = ray.put(artifact_loader)

    # Define training function internally to capture the artifact loader ref as a closure
    # Reference: https://docs.ray.io/en/latest/ray-core/objects.html#closure-capture-of-objects
    def training_function(config_data: dict[str, Any]):
        artifact_loader = ray.get(artifact_loader_ref)
        config = FinetuningJobConfig(**config_data)
        if is_tracking_enabled(config):
            with wandb_init_from_config(
                config.tracking,
                resume=WandbResumeMode.NEVER,
                job_type=LMBuddyJobType.FINETUNING,
            ):
                load_and_train(config, artifact_loader)
        else:
            load_and_train(config, artifact_loader)

    # Construct Ray train configurations from input config
    scaling_config = ScalingConfig(
        use_gpu=config.ray.use_gpu,
        num_workers=config.ray.num_workers,
    )
    run_config = RunConfig(
        name=config.tracking.name if config.tracking else None,
        storage_path=config.ray.storage_path,
        checkpoint_config=CheckpointConfig(num_to_keep=1),
    )
    trainer = TorchTrainer(
        train_loop_per_worker=training_function,
        train_loop_config=config.model_dump(),
        scaling_config=scaling_config,
        run_config=run_config,
    )
    result = trainer.fit()
    print(f"Training result: {result}")

    # Register a model artifact if tracking is enabled and Ray saved a checkpoint
    if config.tracking and result.checkpoint:
        # Must resume from the just-completed training run
        with wandb_init_from_config(config.tracking, resume=WandbResumeMode.MUST) as run:
            model_artifact = build_directory_artifact(
                artifact_name=default_artifact_name(run.name, ArtifactType.MODEL),
                artifact_type=ArtifactType.MODEL,
                dir_path=f"{result.checkpoint.path}/{RayTrainReportCallback.CHECKPOINT_NAME}",
                reference=True,
            )
            print("Logging artifact for model checkpoint...")
            artifact_loader.log_artifact(model_artifact)
